package net.sf.seaf.persistence.util;

import java.beans.PropertyDescriptor;
import java.io.Serializable;
import java.util.IdentityHashMap;
import java.util.Map;

import net.sf.seaf.factory.Factory;
import net.sf.seaf.persistence.CodeTable;
import net.sf.seaf.persistence.CodeTablePropertyRefreshingExceptionProcessor;
import net.sf.seaf.persistence.Entity;
import net.sf.seaf.persistence.ServerController;
import net.sf.seaf.persistence.exception.InvalidPropertyNameException;
import net.sf.seaf.util.bean.PropertyEnumerator;
import net.sf.seaf.util.bean.PropertyVisitor;

import org.apache.commons.beanutils.MethodUtils;
import org.apache.commons.beanutils.PropertyUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Refreshes code table properties of an entity instance.
 * <p>
 * After applied to a transient entity, the refreshed values are managed by the
 * persistent framework and therefore the entity instance can then be saved into
 * the database.
 * <p>
 * This class is thread-safe.
 */
public class CodeTablePropertyRefresher {

	Logger log = LoggerFactory.getLogger(CodeTablePropertyRefresher.class);

	/**
	 * Default constructor.
	 */
	public CodeTablePropertyRefresher() {
	}

	/**
	 * Constructor.
	 * 
	 * @param controllerFactory
	 *            A factory to retrieve seaf controllers from
	 */
	public CodeTablePropertyRefresher(Factory controllerFactory) {
		this.controllerFactory = controllerFactory;
	}

	/**
	 * Refresh all code table properties of an instance of entity, recursing to
	 * all related entities.
	 * 
	 * @param entity
	 *            The instance of an entity
	 */
	public void refresh(Entity<? extends Serializable> entity) {
		new Refresher().refresh(entity);
	}

	/**
	 * Refresh a single code table property of an instance of entity.
	 * 
	 * @param entity
	 *            Entity instance
	 * @param propertyName
	 *            Name of code table property to refresh
	 */
	public void refresh(Entity<? extends Serializable> entity,
			String propertyName) {
		try {
			PropertyDescriptor propertyDescriptor = PropertyUtils
					.getPropertyDescriptor(entity, propertyName);
			if (null == propertyDescriptor) {
				throw new InvalidPropertyNameException(entity, propertyName);
			}
			Class<?> type = propertyDescriptor.getPropertyType();
			Object value = PropertyUtils.getProperty(entity, propertyName);
			if (null != value) {
				refreshProperty(entity, propertyName, type, value);
			}
		} catch (Exception e) {
			getExceptionProcessor()
					.process(entity, null, propertyName, null, e);
		}
	}

	private class Refresher implements PropertyVisitor {
		/**
		 * Keep track of already refreshed instances of entities, to prevent
		 * cyclic recursion.
		 */
		private final Map<Object, Object> refreshedEntities = new IdentityHashMap<Object, Object>();

		private final PropertyEnumerator propertyEnumerator = new PropertyEnumerator();

		private void refresh(Entity<? extends Serializable> entity) {
			refreshedEntities.clear();
			refreshEntityInternal(entity);
		}

		public void visit(Object bean, String name, Class<?> type, Object value) {
			if (CodeTable.class.isAssignableFrom(type) && null != value) {
				// it is a CodeTable property, refresh it
				refreshProperty(bean, name, type, value);
			} else if (Iterable.class.isAssignableFrom(type) && null != value) {
				// it is a property containing a list, refresh each item of the
				// list as if it was an entity
				// NOTE: on seaf entity the list will never be null,
				// but check for it anyway
				Iterable<?> iter = (Iterable<?>) value;
				for (Object item : iter) {
					refreshEntityInternal(item);
				}
			} else {
				// it is a general property, refresh its value as if it was an
				// entity
				refreshEntityInternal(value);
			}
		}

		private void refreshEntityInternal(Object entity) {
			if (null == entity)
				return;
			if (!Entity.class.isAssignableFrom(entity.getClass()))
				return;
			if (refreshedEntities.containsKey(entity)) {
				log.debug("Skipping already refreshed entity {}", entity);
				return;
			}
			log.debug("Refreshing entity {}", entity);
			refreshedEntities.put(entity, null);
			propertyEnumerator.enumerate(entity, this);
		}
	}

	/**
	 * Refresh a property of type CodeTable.
	 * 
	 * @param entity
	 *            Entity
	 * @param name
	 *            Name of the property to refresh
	 * @param type
	 *            Type of the property to refresh
	 * @param value
	 *            Value of the property to refresh
	 */
	private void refreshProperty(Object entity, String name, Class<?> type,
			Object value) {
		try {
			log.trace("Refreshing property {} = {}", name, value);
			Object refreshed = getRefreshedValue(type, value);
			log.trace("Assigning property {} with refreshed value {}", name,
					refreshed);
			PropertyUtils.setSimpleProperty(entity, name, refreshed);
		} catch (Exception e) {
			String code = ((CodeTable) value).getCode();
			getExceptionProcessor().process(entity, type, name, code, e);
		}
	}

	private Object getRefreshedValue(Class<?> type, Object value)
			throws Exception {
		Class<? extends ServerController> controllerType = typeConversionUtil
				.entityToController(type);
		ServerController controller = controllerFactory
				.getInstanceOf(controllerType);
		return MethodUtils
				.invokeExactMethod(controller, "refreshByCode", value);
	}

	private Factory controllerFactory;

	/**
	 * Set the controller factory.
	 * 
	 * @param controllerFactory
	 *            Controller factory
	 */
	public void setControllerFactory(Factory controllerFactory) {
		this.controllerFactory = controllerFactory;
	}

	private TypeConversionUtil typeConversionUtil = TypeConversionUtil
			.getInstance();

	/**
	 * Set the TypeConversionUtil instance.
	 * 
	 * @param typeConversionUtil
	 *            TypeConversionUtil instance
	 */
	public void setTypeConversionUtil(TypeConversionUtil typeConversionUtil) {
		this.typeConversionUtil = typeConversionUtil;
	}

	private CodeTablePropertyRefreshingExceptionProcessor exceptionProcessor;

	/**
	 * Return the exception processor.
	 * 
	 * @return Exception processor
	 */
	private CodeTablePropertyRefreshingExceptionProcessor getExceptionProcessor() {
		if (null == exceptionProcessor) {
			exceptionProcessor = ThrowingCodeTablePropertyRefreshingExceptionProcessor
					.getInstance();
		}
		return exceptionProcessor;
	}

	/**
	 * Set the exception processor.
	 * 
	 * @param exceptionProcessor
	 *            Exception processor
	 */
	public void setExceptionProcessor(
			CodeTablePropertyRefreshingExceptionProcessor exceptionProcessor) {
		this.exceptionProcessor = exceptionProcessor;
	}

}
