package net.sf.seaf.persistence.util;

import java.util.ArrayList;
import java.util.List;

import net.sf.seaf.persistence.Entity;
import net.sf.seaf.persistence.exception.RelationBindingException;
import net.sf.seaf.util.bean.PropertyEnumerator;
import net.sf.seaf.util.bean.PropertyVisitor;

import org.apache.commons.beanutils.PropertyUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Binds relations on beans generated by JBG.
 * <p>
 * Iterates through a tree of JBG beans and re-binds each relation so that both
 * sides are properly set up.
 * <p>
 * This is particularly useful e.g. when a tree is converted by Dozer and then
 * needs to be persisted with Hibernate. Hibernate requires two-way relations to
 * be properly bound in both ways, but Dozer only sets them in one way.
 */
public class RelationBinder {

	private final PropertyEnumerator propertyEnumerator = new PropertyEnumerator();
	private final RelationBindingVisitor visitor = new RelationBindingVisitor();

	/**
	 * Keep track of already refreshed instances of beans, to prevent cyclic
	 * recursion.
	 * <p>
	 * Only skip beans with the same reference; if a bean is equal still process
	 * it.
	 */
	private final List<Object> refreshedBeans = new ArrayList<Object>() {
		@Override
		public boolean contains(Object elem) {
			for (int i = 0; i < size(); i++)
				if (elem == get(i))
					return true;
			return false;
		}

		private static final long serialVersionUID = 1L;
	};

	Logger log = LoggerFactory.getLogger(RelationBinder.class);

	/**
	 * Bind relations on a tree of beans generated by JBG.
	 * <p>
	 * Iterates through a tree of JBG beans and re-binds each relation so that
	 * both sides are properly set up.
	 * 
	 * @param bean
	 *            bean to be bound
	 */
	public void bind(Object bean) {
		refreshedBeans.clear();
		bindInternal(bean);
	}

	private void bindInternal(Object bean) {
		if (null == bean)
			return;
		if (refreshedBeans.contains(bean)) {
			log.debug("Skipping already bound bean {}", bean);
			return;
		}
		log.debug("Binding bean {}", bean);
		refreshedBeans.add(bean);
		propertyEnumerator.enumerate(bean, visitor);
	}

	private class RelationBindingVisitor implements PropertyVisitor {
		public void visit(Object bean, String name, Class<?> type, Object value) {
			String propertyWithReverseSideName = name + "_WithReverseSide";
			if (PropertyUtils.isWriteable(bean, propertyWithReverseSideName)) {
				log.trace("Binding property {}", name);
				try {
					PropertyUtils.setSimpleProperty(bean,
							propertyWithReverseSideName, value);
				} catch (Exception e) {
					String msg = "Could not bind property " + name
							+ " of type " + type + " with value " + value;
					throw new RelationBindingException(msg, e);
				}
			}
			if (Entity.class.isAssignableFrom(type)) {
				bindInternal(value);
			} else if (Iterable.class.isAssignableFrom(type) && null != value) {
				Iterable<?> iter = (Iterable<?>) value;
				for (Object item : iter) {
					bindInternal(item);
				}
			}
		}
	}

}
